#include "./helper.h"

namespace CppUtilities {

/*!
 * \brief Specifies the factor for multiplying all timeouts passed to functions declared in helper.h.
 *
 * The default factor is 1.0.
 */
double timeoutFactor = 1.0;
} // namespace CppUtilities
