#ifndef DEVVIEW_H
#define DEVVIEW_H

#include "./helper.h"

namespace Data {
struct SyncthingDev;
class SyncthingDeviceModel;
class SyncthingSortFilterModel;
} // namespace Data

namespace QtGui {

class DevView : public BasicTreeView {
    Q_OBJECT
public:
    using ModelType = Data::SyncthingDeviceModel;
    using SortFilterModelType = Data::SyncthingSortFilterModel;

    explicit DevView(QWidget *parent = nullptr);

Q_SIGNALS:
    void pauseResumeDev(const Data::SyncthingDev &dev);

protected:
    void mouseReleaseEvent(QMouseEvent *event) override;

private Q_SLOTS:
    void showContextMenu(const QPoint &position);
};
} // namespace QtGui

#endif // DEVVIEW_H
